"""Prepare data for adding techs related to water distribution,
treatment in urban & rural"""

from collections import defaultdict

import pandas as pd
from message_ix import make_df

from message_ix_models.model.water.utils import map_yv_ya_lt
from message_ix_models.util import (
    broadcast,
    make_matched_dfs,
    package_data_path,
    same_node,
    same_time,
)


def add_infrastructure_techs(context):  # noqa: C901
    """Process water distribution data for a scenario instance.
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """
    # TODO reduce complexity of this function from 18 to 15 or less
    # Reference to the water configuration
    info = context["water build info"]

    # define an empty dictionary
    results = {}
    sub_time = context.time
    # load the scenario from context
    scen = context.get_scenario()

    year_wat = [2010, 2015]
    year_wat.extend(info.Y)

    # first activity year for all water technologies is 2020
    first_year = scen.firstmodelyear

    # reading basin_delineation
    FILE2 = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE2)

    df_node = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df_node["node"] = "B" + df_node["BCU_name"].astype(str)
    df_node["mode"] = "M" + df_node["BCU_name"].astype(str)
    if context.type_reg == "country":
        df_node["region"] = context.map_ISO_c[context.regions]
    else:
        df_node["region"] = f"{context.regions}_" + df_node["REGION"].astype(str)

    # Reading water distribution mapping from csv
    path = package_data_path("water", "infrastructure", "water_distribution.xlsx")
    df = pd.read_excel(path)

    techs = [
        "urban_t_d",
        "urban_unconnected",
        "industry_unconnected",
        "rural_t_d",
        "rural_unconnected",
    ]

    df_non_elec = df[df["incmd"] != "electr"].reset_index()
    df_dist = df_non_elec[df_non_elec["tec"].isin(techs)]
    df_non_elec = df_non_elec[~df_non_elec["tec"].isin(techs)]
    df_elec = df[df["incmd"] == "electr"].reset_index()

    inp_df = pd.DataFrame([])

    # Input Dataframe for non elec commodities
    for index, rows in df_non_elec.iterrows():
        inp_df = pd.concat(
            [
                inp_df,
                (
                    make_df(
                        "input",
                        technology=rows["tec"],
                        value=rows["value_mid"],
                        unit="-",
                        level=rows["inlvl"],
                        commodity=rows["incmd"],
                        mode="M1",
                        node_loc=df_node["node"],
                    )
                    .pipe(
                        broadcast,
                        map_yv_ya_lt(
                            year_wat, rows["technical_lifetime_mid"], first_year
                        ),
                        time=sub_time,
                    )
                    .pipe(same_node)
                    .pipe(same_time)
                ),
            ]
        )

    if context.SDG != "baseline":
        for index, rows in df_dist.iterrows():
            inp_df = pd.concat(
                [
                    inp_df,
                    (
                        make_df(
                            "input",
                            technology=rows["tec"],
                            value=rows["value_high"],
                            unit="-",
                            level=rows["inlvl"],
                            commodity=rows["incmd"],
                            mode="Mf",
                        )
                        .pipe(
                            broadcast,
                            map_yv_ya_lt(
                                year_wat, rows["technical_lifetime_mid"], first_year
                            ),
                            node_loc=df_node["node"],
                            time=sub_time,
                        )
                        .pipe(same_node)
                        .pipe(same_time)
                    ),
                ]
            )
    else:
        for index, rows in df_dist.iterrows():
            inp_df = pd.concat(
                [
                    inp_df,
                    (
                        make_df(
                            "input",
                            technology=rows["tec"],
                            value=rows["value_mid"],
                            unit="-",
                            level=rows["inlvl"],
                            commodity=rows["incmd"],
                            mode="M1",
                        )
                        .pipe(
                            broadcast,
                            map_yv_ya_lt(
                                year_wat, rows["technical_lifetime_mid"], first_year
                            ),
                            node_loc=df_node["node"],
                            time=sub_time,
                        )
                        .pipe(same_node)
                        .pipe(same_time)
                    ),
                ]
            )

            inp_df = inp_df.append(
                (
                    make_df(
                        "input",
                        technology=rows["tec"],
                        value=rows["value_high"],
                        unit="-",
                        level=rows["inlvl"],
                        commodity=rows["incmd"],
                        mode="Mf",
                    )
                    .pipe(
                        broadcast,
                        map_yv_ya_lt(
                            year_wat, rows["technical_lifetime_mid"], first_year
                        ),
                        node_loc=df_node["node"],
                        time=sub_time,
                    )
                    .pipe(same_node)
                    .pipe(same_time)
                )
            )
    result_dc = defaultdict(list)

    for index, rows in df_elec.iterrows():
        if rows["tec"] in techs:
            if context.SDG != "baseline":
                inp = make_df(
                    "input",
                    technology=rows["tec"],
                    value=rows["value_high"],
                    unit="-",
                    level="final",
                    commodity="electr",
                    mode="Mf",
                    time_origin="year",
                    node_loc=df_node["node"],
                    node_origin=df_node["region"],
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(
                        year_wat,
                        # 1 because elec commodities don't have technical lifetime
                        1,
                        first_year,
                    ),
                    time=sub_time,
                )

                result_dc["input"].append(inp)
            else:
                inp = make_df(
                    "input",
                    technology=rows["tec"],
                    value=rows["value_high"],
                    unit="-",
                    level="final",
                    commodity="electr",
                    mode="Mf",
                    time_origin="year",
                    node_loc=df_node["node"],
                    node_origin=df_node["region"],
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(
                        year_wat,
                        # 1 because elec commodities don't have technical lifetime
                        1,
                        first_year,
                    ),
                    time=sub_time,
                )

                inp = inp.append(
                    make_df(
                        "input",
                        technology=rows["tec"],
                        value=rows["value_mid"],
                        unit="-",
                        level="final",
                        commodity="electr",
                        mode="M1",
                        time_origin="year",
                        node_loc=df_node["node"],
                        node_origin=df_node["region"],
                    ).pipe(
                        broadcast,
                        # 1 because elec commodities don't have technical lifetime
                        map_yv_ya_lt(year_wat, 1, first_year),
                        time=sub_time,
                    )
                )

                result_dc["input"].append(inp)
        else:
            inp = make_df(
                "input",
                technology=rows["tec"],
                value=rows["value_mid"],
                unit="-",
                level="final",
                commodity="electr",
                mode="M1",
                time_origin="year",
                node_loc=df_node["node"],
                node_origin=df_node["region"],
            ).pipe(
                broadcast,
                map_yv_ya_lt(year_wat, 1, first_year),
                time=sub_time,
            )

            result_dc["input"].append(inp)

    results_new = {par_name: pd.concat(dfs) for par_name, dfs in result_dc.items()}

    inp_df = inp_df.append(results_new["input"])
    # inp_df.dropna(inplace = True)
    results["input"] = inp_df

    # add output dataframe
    df_out = df[~df["outcmd"].isna()]
    df_out_dist = df_out[df_out["tec"].isin(techs)]
    df_out = df_out[~df_out["tec"].isin(techs)]

    out_df = pd.DataFrame([])
    for index, rows in df_out.iterrows():
        out_df = out_df.append(
            (
                make_df(
                    "output",
                    technology=rows["tec"],
                    value=rows["out_value_mid"],
                    unit="-",
                    level=rows["outlvl"],
                    commodity=rows["outcmd"],
                    mode="M1",
                )
                .pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
                .pipe(same_node)
                .pipe(same_time)
            )
        )

    if context.SDG != "baseline":
        out_df = out_df.append(
            make_df(
                "output",
                technology=df_out_dist["tec"],
                value=df_out_dist["out_value_mid"],
                unit="-",
                level=df_out_dist["outlvl"],
                commodity=df_out_dist["outcmd"],
                mode="Mf",
            )
            .pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
            .pipe(same_time)
        )
    else:
        out_df = out_df.append(
            make_df(
                "output",
                technology=df_out_dist["tec"],
                value=df_out_dist["out_value_mid"],
                unit="-",
                level=df_out_dist["outlvl"],
                commodity=df_out_dist["outcmd"],
                mode="M1",
            )
            .pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
            .pipe(same_time)
        )
        out_df = out_df.append(
            make_df(
                "output",
                technology=df_out_dist["tec"],
                value=df_out_dist["out_value_mid"],
                unit="-",
                level=df_out_dist["outlvl"],
                commodity=df_out_dist["outcmd"],
                mode="Mf",
            )
            .pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
            .pipe(same_time)
        )

    results["output"] = out_df

    # Filtering df for capacity factors
    df_cap = df.dropna(subset=["capacity_factor_mid"])
    cap_df = pd.DataFrame([])
    # Adding capacity factor dataframe
    for index, rows in df_cap.iterrows():
        cap_df = cap_df.append(
            make_df(
                "capacity_factor",
                technology=rows["tec"],
                value=rows["capacity_factor_mid"],
                unit="%",
            )
            .pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                node_loc=df_node["node"],
                time=sub_time,
            )
            .pipe(same_node)
        )

    results["capacity_factor"] = cap_df

    # Filtering df for capacity factors
    df_tl = df.dropna(subset=["technical_lifetime_mid"])

    tl = (
        make_df(
            "technical_lifetime",
            technology=df_tl["tec"],
            value=df_tl["technical_lifetime_mid"],
            unit="y",
        )
        .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
        .pipe(same_node)
    )

    results["technical_lifetime"] = tl

    cons_time = make_matched_dfs(tl, construction_time=1)
    results["construction_time"] = cons_time["construction_time"]

    # Investment costs
    df_inv = df.dropna(subset=["investment_mid"])

    # Prepare dataframe for investments
    inv_cost = make_df(
        "inv_cost",
        technology=df_inv["tec"],
        value=df_inv["investment_mid"],
        unit="USD/km3",
    ).pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
    inv_cost = inv_cost[~inv_cost["technology"].isin(techs)]
    results["inv_cost"] = inv_cost

    # Fixed costs
    # Prepare data frame for fix_cost
    fix_cost = pd.DataFrame([])
    var_cost = pd.DataFrame([])

    for index, rows in df_inv.iterrows():
        fix_cost = fix_cost.append(
            make_df(
                "fix_cost",
                technology=df_inv["tec"],
                value=df_inv["fix_cost_mid"],
                unit="USD/km3",
            ).pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                node_loc=df_node["node"],
            )
        )

        fix_cost = fix_cost[~fix_cost["technology"].isin(techs)]

        results["fix_cost"] = fix_cost

    df_var = df_inv[~df_inv["tec"].isin(techs)]
    df_var_dist = df_inv[df_inv["tec"].isin(techs)]

    df_var = df_inv[~df_inv["tec"].isin(techs)]
    df_var_dist = df_inv[df_inv["tec"].isin(techs)]

    if context.SDG != "baseline":
        for index, rows in df_var.iterrows():
            # Variable cost
            var_cost = var_cost.append(
                make_df(
                    "var_cost",
                    technology=rows["tec"],
                    value=rows["var_cost_mid"],
                    unit="USD/km3",
                    mode="M1",
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
            )

        # Variable cost for distribution technologies
        for index, rows in df_var_dist.iterrows():
            var_cost = var_cost.append(
                make_df(
                    "var_cost",
                    technology=rows["tec"],
                    value=rows["var_cost_high"],
                    unit="USD/km3",
                    mode="Mf",
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
            )
        results["var_cost"] = var_cost
    else:
        # Variable cost
        for index, rows in df_var.iterrows():
            var_cost = var_cost.append(
                make_df(
                    "var_cost",
                    technology=rows["tec"],
                    value=df_var["var_cost_mid"],
                    unit="USD/km3",
                    mode="M1",
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
            )

        for index, rows in df_var_dist.iterrows():
            var_cost = var_cost.append(
                make_df(
                    "var_cost",
                    technology=rows["tec"],
                    value=rows["var_cost_mid"],
                    unit="USD/km3",
                    mode="M1",
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
            )

            var_cost = var_cost.append(
                make_df(
                    "var_cost",
                    technology=rows["tec"],
                    value=rows["var_cost_high"],
                    unit="USD/km3",
                    mode="Mf",
                ).pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["technical_lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
            )
        results["var_cost"] = var_cost

    return results


def add_desalination(context):
    """Add desalination infrastructure
    Two types of desalination are considered;
    1. Membrane
    2. Distillation
    Parameters
    ----------
    context : .Context
    Returns
    -------
    data : dict of (str -> pandas.DataFrame)
        Keys are MESSAGE parameter names such as 'input', 'fix_cost'.
        Values are data frames ready for :meth:`~.Scenario.add_par`.
        Years in the data include the model horizon indicated by
        ``context["water build info"]``, plus the additional year 2010.
    """
    # define an empty dictionary
    results = {}
    sub_time = context.time
    # Reference to the water configuration
    info = context["water build info"]

    # load the scenario from context
    scen = context.get_scenario()

    year_wat = [2010, 2015]
    year_wat.extend(info.Y)

    # first activity year for all water technologies is 2020
    first_year = scen.firstmodelyear

    # Reading water distribution mapping from csv
    path = package_data_path("water", "infrastructure", "desalination.xlsx")
    path2 = package_data_path(
        "water",
        "infrastructure",
        f"historical_capacity_desalination_km3_year_{context.regions}.csv",
    )
    path3 = package_data_path(
        "water",
        "infrastructure",
        f"projected_desalination_potential_km3_year_{context.regions}.csv",
    )
    # Reading dataframes
    df_desal = pd.read_excel(path)
    df_hist = pd.read_csv(path2)
    df_proj = pd.read_csv(path3)
    df_proj = df_proj[df_proj["rcp"] == f"{context.RCP}"]
    df_proj = df_proj[~(df_proj["year"] == 2065) & ~(df_proj["year"] == 2075)]
    df_proj.reset_index(inplace=True, drop=True)
    df_proj = df_proj[df_proj["year"].isin(info.Y)]

    # reading basin_delineation
    FILE2 = f"basins_by_region_simpl_{context.regions}.csv"
    PATH = package_data_path("water", "delineation", FILE2)

    df_node = pd.read_csv(PATH)
    # Assigning proper nomenclature
    df_node["node"] = "B" + df_node["BCU_name"].astype(str)
    df_node["mode"] = "M" + df_node["BCU_name"].astype(str)
    if context.type_reg == "country":
        df_node["region"] = context.map_ISO_c[context.regions]
    else:
        df_node["region"] = f"{context.regions}_" + df_node["REGION"].astype(str)
    # output dataframe linking to desal tech types
    out_df = (
        make_df(
            "output",
            technology="extract_salinewater_basin",
            value=1,
            unit="km3/year",
            level="water_avail_basin",
            commodity="salinewater_basin",
            mode="M1",
        )
        .pipe(
            broadcast,
            map_yv_ya_lt(year_wat, 20, first_year),
            node_loc=df_node["node"],
            time=sub_time,
        )
        .pipe(same_node)
        .pipe(same_time)
    )

    tl = (
        make_df(
            "technical_lifetime",
            technology="extract_salinewater_basin",
            value=20,
            unit="y",
        )
        .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
        .pipe(same_node)
    )

    # Historical capacity of desalination technologies
    df_hist_cap = make_df(
        "historical_new_capacity",
        node_loc="B" + df_hist["BCU_name"],
        technology=df_hist["tec_type"],
        year_vtg=df_hist["year"],
        value=df_hist["cap_km3_year"],
        unit="km3/year",
    )
    # Divide the historical capacity by 5 since the existing data is summed over
    # 5 years and model needs per year
    df_hist_cap["value"] = df_hist_cap["value"] / 5

    results["historical_new_capacity"] = df_hist_cap

    # Desalination potentials are added as an upper bound
    # to limit the salinewater extraction
    bound_up = make_df(
        "bound_total_capacity_up",
        node_loc="B" + df_proj["BCU_name"],
        technology="extract_salinewater_basin",
        year_act=df_proj["year"],
        value=df_proj["cap_km3_year"],
        unit="km3/year",
    )
    # Making negative values zero
    bound_up["value"].clip(lower=0, inplace=True)
    # Bound should start from 2025
    bound_up = bound_up[bound_up["year_act"] > 2020]

    results["bound_total_capacity_up"] = bound_up
    # Investment costs
    inv_cost = make_df(
        "inv_cost",
        technology=df_desal["tec"],
        value=df_desal["inv_cost_mid"],
        unit="USD/km3",
    ).pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])

    results["inv_cost"] = inv_cost

    fix_cost = pd.DataFrame([])
    var_cost = pd.DataFrame([])
    for index, rows in df_desal.iterrows():
        # Fixed costs
        # Prepare dataframe for fix_cost
        fix_cost = fix_cost.append(
            make_df(
                "fix_cost",
                technology=rows["tec"],
                value=rows["fix_cost_mid"],
                unit="USD/km3",
            ).pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
                node_loc=df_node["node"],
            )
        )

        results["fix_cost"] = fix_cost

        # Variable cost
        var_cost = var_cost.append(
            make_df(
                "var_cost",
                technology=rows["tec"],
                value=rows["var_cost_mid"],
                unit="USD/km3",
                mode="M1",
            ).pipe(
                broadcast,
                map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
                node_loc=df_node["node"],
                time=sub_time,
            )
        )

    # Dummy  Variable cost for salinewater extrqction
    # var_cost = var_cost.append(
    #     make_df(
    #     "var_cost",
    #     technology='extract_salinewater_basin',
    #     value= 100,
    #     unit="USD/km3",
    #     mode="M1",
    #     time="year",
    # ).pipe(broadcast, year_vtg=year_wat, year_act=year_wat, node_loc=df_node["node"])
    # )

    results["var_cost"] = var_cost

    tl = tl.append(
        (
            make_df(
                "technical_lifetime",
                technology=df_desal["tec"],
                value=df_desal["lifetime_mid"],
                unit="y",
            )
            .pipe(broadcast, year_vtg=year_wat, node_loc=df_node["node"])
            .pipe(same_node)
        )
    )

    results["technical_lifetime"] = tl

    cons_time = make_matched_dfs(tl, construction_time=3)
    results["construction_time"] = cons_time["construction_time"]

    from collections import defaultdict

    result_dc = defaultdict(list)

    for index, rows in df_desal.iterrows():
        inp = make_df(
            "input",
            technology=rows["tec"],
            value=rows["electricity_input_mid"],
            unit="-",
            level="final",
            commodity="electr",
            mode="M1",
            time_origin="year",
            node_loc=df_node["node"],
            node_origin=df_node["region"],
        ).pipe(
            broadcast,
            map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
            time=sub_time,
        )

        result_dc["input"].append(inp)

    results_new = {par_name: pd.concat(dfs) for par_name, dfs in result_dc.items()}

    inp_df = results_new["input"]

    # Adding input dataframe
    df_heat = df_desal[df_desal["heat_input_mid"] > 0]

    result_dc = defaultdict(list)

    for index, rows in df_heat.iterrows():
        inp = make_df(
            "input",
            technology=rows["tec"],
            value=rows["heat_input_mid"],
            unit="-",
            level="final",
            commodity="d_heat",
            mode="M1",
            time_origin="year",
            node_loc=df_node["node"],
            node_origin=df_node["region"],
        ).pipe(
            broadcast,
            map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
            time=sub_time,
        )

        result_dc["input"].append(inp)

    results_new = {par_name: pd.concat(dfs) for par_name, dfs in result_dc.items()}

    inp_df = inp_df.append(results_new["input"])

    # Adding input dataframe
    for index, rows in df_desal.iterrows():
        inp_df = inp_df.append(
            (
                make_df(
                    "input",
                    technology=rows["tec"],
                    value=1,
                    unit="-",
                    level=rows["inlvl"],
                    commodity=rows["incmd"],
                    mode="M1",
                )
                .pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
                .pipe(same_node)
                .pipe(same_time)
            )
        )

        inp_df.dropna(inplace=True)

        results["input"] = inp_df

        out_df = out_df.append(
            (
                make_df(
                    "output",
                    technology=rows["tec"],
                    value=1,
                    unit="-",
                    level=rows["outlvl"],
                    commodity=rows["outcmd"],
                    mode="M1",
                )
                .pipe(
                    broadcast,
                    map_yv_ya_lt(year_wat, rows["lifetime_mid"], first_year),
                    node_loc=df_node["node"],
                    time=sub_time,
                )
                .pipe(same_node)
                .pipe(same_time)
            )
        )

        results["output"] = out_df

    # putting a lower bound on desalination tecs based on hist capacities
    df_bound = df_hist[df_hist["year"] == 2015]
    bound_lo = make_df(
        "bound_activity_lo",
        node_loc="B" + df_bound["BCU_name"],
        technology=df_bound["tec_type"],
        mode="M1",
        value=df_bound["cap_km3_year"],
        unit="km3/year",
    ).pipe(
        broadcast,
        year_act=year_wat,
        time=sub_time,
    )

    bound_lo = bound_lo[bound_lo["year_act"] <= 2030]
    # Divide the histroical capacity by 5 since the existing data is summed over
    # 5 years and model needs per year
    bound_lo["value"] = bound_lo["value"] / 5

    results["bound_activity_lo"] = bound_lo

    return results



import logging
import os
from base64 import b32hexencode
from copy import deepcopy
from pathlib import Path
from random import randbytes
from tempfile import TemporaryDirectory

import message_ix
import pandas as pd
import pytest
from ixmp import config as ixmp_config

from message_ix_models import util
from message_ix_models.util._logging import mark_time
from message_ix_models.util.context import Context

log = logging.getLogger(__name__)

# pytest hooks


def pytest_addoption(parser):
    """Add two command-line options to pytest:

    ``--local-cache``
       Use existing, local cache files in tests. This option can speed up tests that
       *use* the results of slow data loading/parsing. However, if cached values are not
       up to date with the current code, unexpected failure may occur.

    ``--jvmargs``
       Additional arguments to give for the Java Virtual Machine used by :mod:`ixmp`'s
       :class:`.JDBCBackend`. Used by :func:`session_context`.
    """
    parser.addoption(
        "--local-cache",
        action="store_true",
        help="Use existing local cache files in tests",
    )
    parser.addoption(
        "--jvmargs",
        action="store",
        default="",
        help="Arguments for Java VM used by ixmp JDBCBackend",
    )


# Fixtures


@pytest.fixture(scope="session")
def session_context(pytestconfig, tmp_env):
    """A :class:`.Context` connected to a temporary, in-memory database.

    This Context is suitable for modifying and running test code that does not affect
    the user/developer's filesystem and configured :mod:`ixmp` databases.

    Uses the :func:`.tmp_env` fixture from ixmp. This fixture also sets:

    - :attr:`.Context.cache_path`, depending on whether the :program:`--local-cache` CLI
      option was given:

      - If not given: pytest's :doc:`standard cache directory <pytest:how-to/cache>`.
      - If given: the :file:`/cache/` directory under the user's "message local data"
        directory.

    - the "message local data" config key to a temporary directory :file:`/data/` under
      the :ref:`pytest tmp_path directory <pytest:tmp_path>`.

    """
    from platformdirs import user_cache_path

    ctx = Context.only()

    # Temporary, empty local directory for local data
    session_tmp_dir = Path(pytestconfig._tmp_path_factory.mktemp("data"))

    # Set the cache path according to whether pytest --local-cache was given. If True,
    # pick up the existing setting from the user environment. If False, use a pytest-
    # managed cache directory that persists across test sessions.
    ctx.cache_path = (
        user_cache_path("message-ix-models", ensure_exists=True)
        if pytestconfig.option.local_cache
        # TODO use pytestconfig.cache.mkdir() when pytest >= 6.3 is available
        else Path(pytestconfig.cache.makedir("cache"))
    )

    # Other local data in the temporary directory for this session only
    ctx.local_data = session_tmp_dir

    # Also set the "message local data" key in the ixmp config
    ixmp_config.set("message local data", session_tmp_dir)

    # If message_data is not installed, use a temporary path for private_data_path()
    message_data_path = util.MESSAGE_DATA_PATH
    if util.MESSAGE_DATA_PATH is None:
        util.MESSAGE_DATA_PATH = session_tmp_dir.joinpath("message_data")

        # Create some subdirectories
        util.MESSAGE_DATA_PATH.joinpath("data", "tests").mkdir(parents=True)

    # Add a platform connected to an in-memory database
    platform_name = "message-ix-models"
    ixmp_config.add_platform(
        platform_name,
        "jdbc",
        "hsqldb",
        url=f"jdbc:hsqldb:mem://{platform_name}",
        jvmargs=pytestconfig.option.jvmargs,
    )
    ixmp_config.save()

    ctx.platform_info["name"] = platform_name

    try:
        yield ctx
    finally:
        ctx.close_db()
        ixmp_config.remove_platform(platform_name)

        # Restore prior value
        util.MESSAGE_DATA_PATH = message_data_path


@pytest.fixture(scope="function")
def test_context(request, session_context):
    """A copy of :func:`session_context` scoped to one test function."""
    ctx = deepcopy(session_context)

    # Ensure there is a report key
    ctx.setdefault("report", dict())

    yield ctx

    ctx.delete()


@pytest.fixture(scope="function")
def user_context(request):  # pragma: no cover
    """Context which can access user's configuration, e.g. platform names."""
    # Disabled; this is bad practice
    raise NotImplementedError


@pytest.fixture
def mix_models_cli(session_context, tmp_env):
    """A :class:`.CliRunner` object that invokes the :program:`mix-models` CLI.

    NB this requires:

    - The :mod:`ixmp` :func:`.tmp_env` fixture. This sets ``IXMP_DATA`` to a temporary
      directory managed by :mod:`pytest`.
    - The :func:`session_context` fixture. This (a) sets :attr:`.Config.local_data` to
      a temporary directory within ``IXMP_DATA`` and (b) ensures changes to
      :class:`.Context` made by invoked commands do not reach other tests.
    """
    from message_ix_models import cli
    from message_ix_models.util.click import CliRunner

    yield CliRunner(cli.main, cli.__name__, env=tmp_env)


# Testing utility functions


def bare_res(request, context: Context, solved: bool = False) -> message_ix.Scenario:
    """Return or create a |Scenario| containing the bare RES, for use in testing.

    The Scenario has a model name like "MESSAGEix-GLOBIOM [regions] Y[years]", for
    instance "MESSAGEix-GLOBIOM R14 YB" (see :func:`.bare.name`) and a scenario name
    either from :py:`request.node.name` or "baseline" plus a random string.

    This function should:

    - only be called from within test code, i.e. in :mod:`message_data.tests`.
    - be called once for each test function, so that each test receives a fresh copy of
      the RES scenario.

    Parameters
    ----------
    request : .Request or None
        The pytest :fixture:`pytest:request` fixture. If provided the pytest test node
        name is used for the scenario name of the returned Scenario.
    context : Context
        Passed to :func:`.testing.bare_res`.
    solved : bool, *optional*
        Return a solved Scenario.

    Returns
    -------
    Scenario
        The scenario is a fresh clone, so can be modified freely without disturbing
        other tests.
    """
    from message_ix_models.model import bare

    name = bare.name(context)
    mp = context.get_platform()

    try:
        base = message_ix.Scenario(mp, name, "baseline")
    except ValueError:
        log.info(f"Create '{name}/baseline' for testing")
        context.scenario_info.update(model=name, scenario="baseline")
        base = bare.create_res(context)

    if solved and not base.has_solution():
        log.info("Solve")
        base.solve(solve_options=dict(lpmethod=4), quiet=True)

    try:
        new_name = request.node.name
    except AttributeError:
        # Generate a new scenario name with a random part
        new_name = f"baseline {b32hexencode(randbytes(3)).decode().rstrip('=').lower()}"

    log.info(f"Clone to '{name}/{new_name}'")
    return base.clone(scenario=new_name, keep_solution=solved)


#: Items with names that match (partially or fully) these names are omitted by
#: :func:`export_test_data`.
EXPORT_OMIT = [
    "aeei",
    "cost_MESSAGE",
    "demand_MESSAGE",
    "demand",
    "depr",
    "esub",
    "gdp_calibrate",
    "grow",
    "historical_gdp",
    "kgdp",
    "kpvs",
    "lakl",
    "land",
    "lotol",
    "mapping_macro_sector",
    "MERtoPPP",
    "prfconst",
    "price_MESSAGE",
    "ref_",
    "sector",
]


def export_test_data(context: Context):
    """Export a subset of data from a scenario, for use in tests.

    The context settings ``export_nodes`` (default: "R11_AFR" and "R11_CPA") and
    ``export_techs`` (default: "coal_ppl") are used to filter the data exported.
    In addition, any item (set, parameter, variable, or equation) with a name matching
    :data:`EXPORT_OMIT` *or* the context setting ``export_exclude`` is discarded.

    The output is stored at :file:`data/tests/{model name}_{scenario name}_{techs}.xlsx`
    in :mod:`message_data`.

    See also
    --------
    :ref:`export-test-data`
    """
    from message_ix_models.util import private_data_path

    # Load the scenario to be exported
    scen = context.get_scenario()

    # Retrieve the context settings giving the nodes and technologies to export
    nodes = context.get("export_nodes", ["R11_AFR", "R11_CPA"])
    technology = context.get("export_techs", ["coal_ppl"])

    # Construct the destination file name
    dest_file = private_data_path(
        "tests", f"{scen.model}_{scen.scenario}_{'_'.join(technology)}.xlsx"
    )
    # Temporary file name
    td = TemporaryDirectory()
    tmp_file = Path(td.name).joinpath("export_test_data.xlsx")

    # Ensure the target directory exists
    dest_file.parent.mkdir(exist_ok=True)

    # Dump data to temporary Excel file
    log.info(f"Export test data to {dest_file}")
    scen.to_excel(
        tmp_file,
        filters={
            "technology": technology,
            "node": nodes,
            "node_dest": nodes,
            "node_loc": nodes,
            "node_origin": nodes,
            "node_parent": nodes,
            "node_rel": nodes,
            "node_share": nodes,
        },
    )

    mark_time()

    log.info("Reduce test data")

    # Read from temporary file and write to final file, omitting unnecessary sheets
    reader = pd.ExcelFile(tmp_file)
    writer = pd.ExcelWriter(dest_file)

    # Retrieve the type mapping first, to be modified as sheets are discarded
    ix_type_mapping = reader.parse("ix_type_mapping").set_index("item")

    for name in reader.sheet_names:
        # Check if this sheet is to be included
        if name == "ix_type_mapping":
            # Already handled
            continue
        elif any(i in name for i in (EXPORT_OMIT + context.get("export_exclude", []))):
            log.info(f"Discard sheet '{name}'")

            # Remove from the mapping
            ix_type_mapping.drop(name, inplace=True)

            continue

        # Copy the sheet from temporary to final file
        reader.parse(name).to_excel(writer, sheet_name=name, index=False)

    # Close the temporary file
    reader.close()

    # Write the mapping
    ix_type_mapping.reset_index().to_excel(
        writer, sheet_name="ix_type_mapping", index=False
    )

    # Close the final file
    writer.close()

    mark_time()


#: Shorthand for marking a parametrized test case that is expected to fail because it is
#: not implemented.
NIE = pytest.mark.xfail(raises=NotImplementedError)

#: :data:`True` if tests occur on GitHub Actions.
GHA = "GITHUB_ACTIONS" in os.environ


def not_ci(reason=None, action="skip"):
    """Mark a test as xfail or skipif if on CI infrastructure.

    Checks the ``GITHUB_ACTIONS`` environment variable; returns a pytest mark.
    """
    action = "skipif" if action == "skip" else action
    return getattr(pytest.mark, action)(condition=GHA, reason=reason)



import logging

from genno import Computer
from ixmp.testing import assert_logs

from message_ix_models import ScenarioInfo
from message_ix_models.model.structure import get_codes
from message_ix_models.report import prepare_reporter
from message_ix_models.report.compat import (
    TECH_FILTERS,
    callback,
    get_techs,
    prepare_techs,
)
from message_ix_models.report.sim import to_simulate

from ..test_report import ss_reporter


@to_simulate.minimum_version
def test_compat(tmp_path, test_context):
    import numpy.testing as npt

    rep = ss_reporter()
    prepare_reporter(test_context, reporter=rep)

    rep.add("scenario", ScenarioInfo(model="Model name", scenario="Scenario name"))

    # Tasks can be added to the reporter
    callback(rep, test_context)

    # Select a key
    key = (
        "transport emissions full::iamc"  # IAMC structure
        # "Transport"  # Top level
        # "Hydrogen_trp"  # Second level
        # "inp_nonccs_gas_tecs_wo_CCSRETRO"  # Third level
        # "_26"  # Fourth level
    )

    # commented: Show what would be done
    # print(rep.describe(key))
    # rep.visualize(tmp_path.joinpath("visualize.svg"), key)

    # Calculation runs
    result = rep.get(key)

    # print(result.to_string())  # For intermediate results

    df = result.as_pandas()  # For pyam.IamDataFrame, which doesn't have .to_string()

    # commented: Display or save output
    # print(df.to_string())
    # df.to_csv(tmp_path.joinpath("transport-emissions-full.csv"), index=False)

    # Check a specific value
    # TODO Expand set of expected values
    npt.assert_allclose(
        df.query("region == 'R11_AFR' and year == 2020")["value"].item(), 54.0532
    )


def test_prepare_techs(test_context):
    from message_ix_models.model.bare import get_spec

    # Retrieve a spec with the default set of technologies
    spec = get_spec(test_context)
    technologies = spec.add.set["technology"]

    c = Computer()
    prepare_techs(c, technologies)

    # Expected sets of technologies based on the default technology.yaml
    assert {
        "gas all": [
            "gas_cc",
            "gas_ct",
            "gas_fs",
            "gas_hpl",
            "gas_htfc",
            "gas_i",
            "gas_ppl",
            "gas_rc",
            "gas_t_d",
            "gas_t_d_ch4",
            "gas_trp",
            "hp_gas_i",
            "hp_gas_rc",
        ],
        "gas extra": [],
        # Residential and commercial
        "rc gas": ["gas_rc", "hp_gas_rc"],
        # Transport
        "trp coal": ["coal_trp"],
        "trp foil": ["foil_trp"],
        "trp gas": ["gas_trp"],
        "trp loil": ["loil_trp"],
        "trp meth": ["meth_fc_trp", "meth_ic_trp"],
    } == {k: get_techs(c, k) for k in TECH_FILTERS}


def test_prepare_techs_filter_error(caplog, monkeypatch):
    """:func:`.prepare_techs` logs warnings for invalid filters."""
    monkeypatch.setitem(TECH_FILTERS, "foo", "not a filter")

    with assert_logs(caplog, "SyntaxError('invalid syntax", at_level=logging.WARNING):
        prepare_techs(Computer(), get_codes("technology"))



from typing import Dict

try:
    # ixmp 3.8.0 and later
    from ixmp.util import (
        discard_on_error,
        maybe_check_out,
        maybe_commit,
        parse_url,
        show_versions,
    )
except ImportError:
    # ixmp <= 3.7.0
    from contextlib import nullcontext

    from ixmp.utils import (  # type: ignore [import-not-found,no-redef]  # noqa: F401
        maybe_check_out,
        maybe_commit,
        parse_url,
        show_versions,
    )

    def discard_on_error(*args):
        return nullcontext()


def rename_dims() -> Dict[str, str]:
    """Access :data:`.ixmp.report.common.RENAME_DIMS`.

    This provides backwards-compatibility with ixmp versions 3.7.0 and earlier. It can
    be removed when message-ix-models no longer supports versions of ixmp older than
    3.8.0.
    """
    try:
        # ixmp 3.8.0 and later
        import ixmp.report.common
    except ImportError:
        # ixmp <= 3.7.0
        import ixmp.reporting.util  # type: ignore [import-not-found]

        return ixmp.reporting.util.RENAME_DIMS
    else:
        return ixmp.report.common.RENAME_DIMS



import logging
import re
from typing import Optional, Tuple

import pandas as pd
from genno import Quantity
from genno import operator as g
from iam_units import convert_gwp
from message_ix import Scenario, make_df

from message_ix_models import ScenarioInfo
from message_ix_models.util import package_data_path

from .structure import get_codes

log = logging.getLogger(__name__)


def get_emission_factors(units: Optional[str] = None) -> Quantity:
    """Return carbon emission factors.

    Values are from the file :file:`message_ix_models/data/ipcc/1996_v3_t1-2.csv`, in
    turn from `IPCC <https://www.ipcc-nggip.iges.or.jp/public/gl/guidelin/ch1wb1.pdf>`_
    (see Table 1-2 on page 1.6); these are the same that appear on the "Emissions from
    energy" page of the MESSAGEix-GLOBIOM documentation.

    The fuel dimension and names in the source are mapped to a :math:`c` ("commodity")
    dimension and labels from :ref:`commodity.yaml <commodity-yaml>`, using the
    ``ipcc-1996-name`` annotations appearing in the latter. A value for "methanol" that
    appears in the MESSAGEix-GLOBIOM docs table but not in the source is appended.

    Parameters
    ----------
    unit : str, *optional*
        Expression for units of the returned quantity. Tested values include:

        - “tC / TJ”, source units (default),
        - “t CO2 / TJ”, and
        - “t C / kWa”, internal units in MESSAGEix-GLOBIOM, for instance for
          "relation_activity" entries for emissions relations.

    Returns
    -------
    Quantity
        with 1 dimension (:math:`c`).
    """
    # Prepare information about commodities
    commodities = get_codes("commodity")
    relabel = {}  # Mapping from IPCC names/IDs to message_ix_models commodity ID
    select = []  # Select only the commodities needed
    for c in commodities:
        try:
            ipcc_name = str(c.get_annotation(id="ipcc-1996-name").text)
        except KeyError:
            continue
        else:
            relabel[ipcc_name] = c.id
            select.append(c.id)

    # Load data from file; relabel; and select only the values needed
    result = (
        g.load_file(package_data_path("ipcc", "1996_v3_t1-2.csv"), dims={"fuel": "c"})
        .pipe(g.relabel, dict(c=relabel))
        .pipe(g.select, dict(c=select))
    )

    # Manually insert a value for methanol
    result = g.concat(
        result,
        Quantity(pd.Series(17.4, pd.Index(["methanol"], name="c")), units=result.units),
    )

    result.attrs["species"] = "C"

    if units is not None:
        # Identify a GWP factor for target `units`, if any
        to_units, to_species = split_species(units)
        gwp_factor = convert_gwp(
            "AR5GWP100", (1.0, str(result.units)), "C", to_species
        ).magnitude
    else:
        gwp_factor, to_units = 1.0, result.units

    # Multiply by the GWP factor; let genno/pint handle other conversion
    return result.pipe(g.mul, Quantity(gwp_factor)).pipe(g.convert_units, to_units)


def add_tax_emission(
    scen: Scenario,
    price: float,
    conversion_factor: Optional[float] = None,
    drate_parameter="drate",
) -> None:
    """Add a global CO₂ price to `scen`.

    A carbon price is implemented on node=“World” by populating the
    :ref:`MESSAGEix parameter <message-ix:section_parameter_emissions>`
    ``tax_emission``, starting from the first model year and covering the entire model
    horizon. The tax has an annual growth rate equal to the discount rate.

    The other dimensions of ``tax_emission`` are filled with type_emission=“TCE” and
    type_tec=“all”.

    Parameters
    ----------
    scen : :class:`message_ix.Scenario`
    price : float
        Price in the first model year, in USD / tonne CO₂.
    conversion_factor : float, *optional*
        Factor for converting `price` into the model's internal emissions units,
        currently USD / tonne carbon. Optional: a default value is retrieved from
        :mod:`iam_units`.
    drate_parameter : str; one of "drate" or "interestrate"
        Name of the parameter to use for the growth rate of the carbon price.
    """
    years = ScenarioInfo(scen).Y
    filters = dict(year=years)
    # Default: since the mass of the species is in the denominator, take the inverse
    conversion_factor = conversion_factor or 1.0 / convert_gwp(
        "AR5GWP100", "1 t", "CO2", "C"
    )

    # Duration of periods
    dp = scen.par("duration_period", filters=filters).set_index("year")["value"]

    # Retrieve the discount rate
    if drate_parameter == "interestrate":
        # MESSAGE parameter with "year" dimension
        r = scen.par(drate_parameter, filters=filters).set_index("year")["value"]
    else:
        # MACRO parameter with "node" dimension
        drates = scen.par(drate_parameter).value.unique()
        if len(drates) > 1:
            log.warning(f"Using the first of multiple discount rates: drate={drates}")
        r = pd.Series([drates[0]] * len(years), index=pd.Index(years, name="year"))

    # Compute cumulative growth versus the first period
    r_cumulative = (r + 1).pow(dp.shift(-1)).cumprod().shift(1, fill_value=1.0)

    # Assemble the parameter data
    name = "tax_emission"
    data = make_df(
        name,
        value=(price * conversion_factor * r_cumulative),
        type_year=r_cumulative.index,
        node="World",
        type_emission="TCE",
        type_tec="all",
        unit="USD/tC",
    )

    with scen.transact("Added carbon price"):
        scen.add_par(name, data)


def split_species(unit_expr: str) -> Tuple[str, Optional[str]]:
    """Split `unit_expr` to an expression without a unit mention, and maybe species."""
    if match := re.fullmatch("(.*)(CO2|C)(.*)", unit_expr):
        return f"{match.group(1)}{match.group(3)}", match.group(2)
    else:
        return unit_expr, None
